﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;

namespace Pololu
{
    /// <summary>
    /// A CSharp wrapper class for all the sloscope DLL functions.
    /// </summary>
    public class Sloscope : IDisposable
    {
        IntPtr handle;

        public static Sloscope connect()
        {
            IntPtr handle = sloscopeConnect();
            checkErrors();
            return new Sloscope(handle);
        }

        public void Dispose()
        {
            sloscopeFree(handle);
            handle = IntPtr.Zero;
        }

        public UInt16 getReferenceVoltage()
        {
            UInt16 referenceVoltage = sloscopeGetReferenceVoltage(handle);
            checkErrors();
            return referenceVoltage;
        }

        public String serialNumber
        {
            get
            {
                String serialNumber = sloscopeGetSerialNumber(handle);
                checkErrors();
                return serialNumber;
            }
        }

        public void flushBuffers()
        {
            sloscopeFlushBuffers(handle);
            checkErrors();
        }

        public State state
        {
            set
            {
                sloscopeSetState(handle, (byte)value);
                checkErrors();
            }
            get
            {
                State s = (State)sloscopeGetState(handle);
                checkErrors();
                return s;
            }
        }

        public UInt16 period
        {
            set
            {
                sloscopeSetPeriod(handle, value);
                checkErrors();
            }
            get
            {
                UInt16 period = sloscopeGetPeriod(handle);
                checkErrors();
                return period;
            }

        }

        public OutputState channelAOutputState
        {
            set
            {
                setOutputState(value, OutputState.NoChange);
            }
            get
            {
                return getOutputState()[0];
            }
        }

        public OutputState channelBOutputState
        {
            set
            {
                setOutputState(OutputState.NoChange, value);
            }
            get
            {
                return getOutputState()[1];
            }
        }

        public void setOutputState(OutputState channelAOutputState, OutputState channelBOutputState)
        {
            sloscopeSetOutputState(handle, (byte)channelAOutputState, (byte)channelBOutputState);
            checkErrors();
        }

        public OutputState[] getOutputState()
        {
            UInt16 outputStateInt = sloscopeGetOutputState(handle);
            checkErrors();
            return new OutputState[] { (OutputState)(outputStateInt & 0xFF), (OutputState)(outputStateInt >> 8 & 0xFF) }; 
        }

        public UInt32 checkMissedReadings()
        {
            return sloscopeCheckMissedReadings(handle);
        }

        public UInt32 checkMissedFrames()
        {
            return sloscopeCheckMissedFrames(handle);
        }

        public UInt32 read(IList<Reading> buffer, ReadMode readMode, UInt32 offset, UInt32 length, UInt32 timeoutMS)
        {
            // This ugliness is only necessary because David can not figure out how to pass
            // an array of structs to unmanaged code and have the unmanaged code modify those structs,
            // without adding another layer of indirection using (ref Reading[] buffer).

            Byte[] tmpBuffer = new Byte[length * 2];
            UInt32 readingCount = sloscopeRead(handle, tmpBuffer, length*2, (byte)readMode, timeoutMS)/2;

            checkErrors();

            for (UInt32 i = 0; i < readingCount; i++)
            {
                Sloscope.Reading tmpReading;
                tmpReading.channelA = tmpBuffer[2 * i];
                tmpReading.channelB = tmpBuffer[2 * i + 1];
                
                buffer[(int)(offset + i)] = tmpReading;
            }
            return readingCount;
        }

        public enum State
        {
            Off = 0,
            TwoAnalog = 1,
            OneAnalogOneDigital = 2
        }

        public enum ReadMode
        {
            FillBuffer = 0,
            Existing = 1
        }

        public enum OutputState
        {
            Off = 0,
            Low = 1,
            High = 3,
            NoChange = 0xFF
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct Reading
        {
            public Byte channelA;
            public Byte channelB;
        }

        Sloscope(IntPtr handle)
        {
            this.handle = handle;
        }

        static void checkErrors()
        {
            switch (Marshal.GetLastWin32Error())
            {
                case 0:
                    return;

                case 0x20020103:
                    throw new Exception("No sloscope was found.  Check the Device Manager and make sure that a sloscope is plugged in to the computer and the drivers are installed.");

                case 0x20030103:
                    throw new Exception("No sloscope was found.  Check the Device Manager and make sure that a sloscope is plugged in to the computer and the drivers are installed.");

                case 0x20050005:
                    throw new Exception("Access was denied when trying to open the sloscope.  Another program may be using the sloscope.  Try closing all Sloscope clients and all Pololu USB device configurators, and try unplugging and plugging the device in.");

                case 0x21010000:
                    throw new Exception("You must call setState() with a value other than Sloscope.State.Off before calling read().");

                case 0x3E4:
                case 0x2101001F:
                case 0x210103E3:
                case 0x2102001F:
                case 0x210203E3:
                case 0x2103001F:
                case 0x210303E3:
                    throw new Exception("Communication with the sloscope failed.  Try reconnecting the programmer to USB.  Error code 0x" + Marshal.GetLastWin32Error().ToString("x") + ".");

                default:
                    throw new Exception("Unexpected error (0x" + Marshal.GetLastWin32Error().ToString("x") + ").  Try cycling programmer power.");
            }
        }

        [DllImport("sloscope.dll", SetLastError = true)]
        static extern IntPtr sloscopeConnect();

        [DllImport("sloscope.dll", SetLastError = true)]
        static extern void sloscopeFree(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError = true, CharSet=CharSet.Unicode)]
        static extern String sloscopeGetSerialNumber(IntPtr listHandle);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern void sloscopeSetState(IntPtr scope, Byte state);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern Byte sloscopeGetState(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern UInt16 sloscopeGetReferenceVoltage(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern UInt32 sloscopeRead(IntPtr scope, Byte[] buffer, UInt32 bufferLength, Byte readMode, UInt32 timeoutMS);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern void sloscopeFlushBuffers(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern UInt32 sloscopeCheckMissedReadings(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern UInt32 sloscopeCheckMissedFrames(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern void sloscopeSetPeriod(IntPtr scope, UInt16 period);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern UInt16 sloscopeGetPeriod(IntPtr scope);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern void sloscopeSetOutputState(IntPtr scope, Byte channelAOutputState, Byte channelBOutputState);

        [DllImport("sloscope.dll", SetLastError=true)]
        static extern UInt16 sloscopeGetOutputState(IntPtr scope);
        
        [DllImport("sloscope.dll", SetLastError=true, EntryPoint="sloscopeListCreate")]
        public static extern IntPtr listCreate();

        [DllImport("sloscope.dll", SetLastError = true, EntryPoint = "sloscopeListGetSerialNumber", CharSet = CharSet.Unicode)]
        public static extern String listGetSerialNumber(IntPtr listHandle, Byte index);

        [DllImport("sloscope.dll", SetLastError=true)]
        private static extern IntPtr listConnect(IntPtr listHandle, Byte index);

        [DllImport("sloscope.dll", SetLastError=true, EntryPoint="sloscopeListDestroy")]
        public static extern void listDestroy(IntPtr listHandle);

    }
}
